'use strict';
var _ = require('underscore');
var path = require('path');
var gutil = require('gulp-util');
var through = require('through2');
var objectAssign = require('object-assign');
var file = require('vinyl-file');
var sortKeys = require('sort-keys');
var vfs = require('vinyl-fs');

var defaults = {
	path: 'rev-mapping.json',
	// Apply the default JSON transformer.
	// The user can pass in his on transformer if he wants. The only requirement is that it should
	// support 'parse' and 'stringify' methods.
	transformer: JSON,
    revSuffix: '-[0-9a-f]{8,10}-?'
};

var mappingSuffix = /([^\"\'\\\/]*\.(js|css|png|jpg|gif))/gm;

Array.prototype.unique = function() {
    var n = {}
    ,   r = [];

    this.forEach(function(v) {
        if (!n[v]) {
            n[v] = true
            r.push(v)
        }
    });
    return r;
};

function _getManifestData(file, opts) {
    var data;
    var ext = path.extname(file.path);
    if (ext === '.json') {
        var json = {};
        try {
            var content = file.contents.toString('utf8');
            if (content) {
                json = JSON.parse(content);
            }
        } catch (x) {
            this.emit('error', new PluginError(PLUGIN_NAME,  x));
            return;
        }
        if (_.isObject(json)) {
            var jsonType;
            if(json['this_is_mapping_file']) {
            	jsonType = 'map';
            }
            if(!jsonType) {
            	Object.keys(json).forEach(function (key) {
	                if ( !_.isString(json[key]) || path.basename(json[key]).replace(new RegExp( opts.revSuffix ), '' ) !==  path.basename(key) ) {
	                    jsonType = 'json';
	                    return;
	                }
	            });
            }

            if(!jsonType) {
            	jsonType = 'rev';
            }

            if (jsonType != 'json') {
                data = {
                	type: jsonType,
                	data: json
                };
            }
        }

    }
    return data;
}

var plugin = function (pth, opts) {
	opts = objectAssign(defaults, opts, pth);

	var sourcemap = [];

	var manifest = {
		this_is_mapping_file: true
	};

	return through.obj(function (file, enc, cb) {
		if (file.isNull()) {
			cb(null, file);
			return;
		}

		if (file.isStream()) {
			cb(new gutil.PluginError('gulp-rev-mapping', 'Streaming not supported'));
			return;
		}

		var relPath = path.relative(process.cwd(), file.path).replace(/\\/g, '/');
		var fileList = file.contents.toString().match(mappingSuffix);

		if(fileList) {
			fileList = fileList.map(function(filename) {
				return filename.replace(new RegExp( opts.revSuffix ), '');
			});
			fileList = fileList.unique();
		}

		var mData = _getManifestData.call(this, file, opts);
		if (mData && mData.type == 'map') {
            _.extend( manifest, mData.data );
        }

        if (!mData) {
        	var tmp = {relPath: relPath, fileList: fileList};
        	sourcemap.push(tmp);
        }

		cb();
	}, function (cb) {
		var file = new gutil.File(opts);
		sourcemap.forEach(function(file) {
			if(manifest[file.relPath]) {
				manifest[file.relPath].forEach(function(oldFile) {
					if(!file.fileList || file.fileList.indexOf(oldFile) == -1) {
						manifest[file.relPath].splice(manifest[file.relPath].indexOf(oldFile), 1);
						if(manifest[oldFile]) {
							manifest[oldFile].splice(manifest[oldFile].indexOf(file.relPath, 1))
						}
						if(manifest[file.relPath].length == 0) {
							delete manifest[file.relPath];
						}
						if(manifest[oldFile] && manifest[oldFile].length == 0) {
							delete manifest[oldFile];
						}
					} else {
						if(manifest[file.relPath].indexOf(oldFile) == -1) {
							manifest[file.relPath].push(oldFile);
						}
						if(manifest[oldFile].indexOf(file.relPath) == -1) {
							manifest[oldFile].push(file.relPath);
						}
					}
				});
			} else {
				if(file.fileList) {
					manifest[file.relPath] = file.fileList;
					file.fileList.forEach(function(filename) {
						if(!manifest[filename]) {
							manifest[filename] = [];
						}
						if(manifest[filename].indexOf(file.relPath) == -1) {
							manifest[filename].push(file.relPath);
						}
					});
				}
			}
		});

		file.contents = new Buffer(opts.transformer.stringify(sortKeys(manifest), null, '  '));
		this.push(file);
		cb();;
	});
};

plugin.replace = function(replaceOne, opts) {
	opts = objectAssign(defaults, opts);
	var map = {};
	var manifest = {};
	var sourcemap = [];
	return through.obj(function (file, enc, cb) {
		if (file.isNull()) {
			cb(null, file);
			return;
		}

		if (file.isStream()) {
			cb(new gutil.PluginError('gulp-rev-mapping', 'Streaming not supported'));
			return;
		}

		file.relPath = path.relative(process.cwd(), file.path).replace(/\\/g, '/');

		var mData = _getManifestData.call(this, file, opts);
		if (mData && mData.type == 'map') {
            _.extend( map, mData.data );
        }
        if (mData && mData.type == 'rev') {
            _.extend( manifest, mData.data );
        }
        if(!mData) {
            sourcemap.push(file);
        }
		cb();
	}, function(cb) {
		sourcemap.forEach(function (file){
            if (!file.isNull()) {
            	var src = file.contents.toString('utf8');
            	if(replaceOne) {
            		var newReplace = path.basename(replaceOne).replace(new RegExp( opts.revSuffix ), '' );
            		if(!manifest[newReplace] || !map[newReplace] || map[newReplace].indexOf(file.relPath) == -1) {
            			return;
            		}
            		var pattern = 
                		path.basename(newReplace, path.extname(newReplace))
                		+ '('
                		+ opts.revSuffix
                		+ ')?'
                    	+ path.extname(newReplace);
            		src = src.replace(new RegExp(pattern, 'gm'), manifest[newReplace]);
            	} else {
	                if(map[file.relPath]) {
		                map[file.relPath].forEach(function (r) {
		                	var pattern = 
		                		path.basename(r, path.extname(r))
		                		+ '('
		                		+ opts.revSuffix
		                		+ ')?'
                            	+ path.extname(r);
		                	var reg = new RegExp(pattern, 'gm');
		                	if(manifest[r]) {
		                    	src = src.replace(reg, manifest[r]);
		                    }
		                });
	                }
                }
                if(file.contents.toString() != src.toString()) {
	                file.contents = new Buffer(src);
	                this.push(file);
                }
            }
        }, this);
        cb();
	});
};

module.exports = plugin;
